package org.antlr.tool;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.antlr.analysis.DFAState;
import org.antlr.analysis.RuleClosureTransition;
import org.antlr.analysis.SemanticContext;
import org.antlr.analysis.State;
import org.antlr.analysis.Transition;
import org.antlr.misc.Utils;
import org.antlr.tool.ErrorManager;
import org.antlr.tool.Grammar;

public class FASerializer {
  protected Set<State> markedStates;
  
  protected int stateCounter = 0;
  
  protected Map<State, Integer> stateNumberTranslator;
  
  protected Grammar grammar;
  
  public FASerializer(Grammar grammar) {
    this.grammar = grammar;
  }
  
  public String serialize(State s) {
    if (s == null)
      return "<no automaton>"; 
    return serialize(s, true);
  }
  
  public String serialize(State s, boolean renumber) {
    this.markedStates = new HashSet<State>();
    this.stateCounter = 0;
    if (renumber) {
      this.stateNumberTranslator = new HashMap<State, Integer>();
      walkFANormalizingStateNumbers(s);
    } 
    List<String> lines = new ArrayList<String>();
    if (s.getNumberOfTransitions() > 0) {
      walkSerializingFA(lines, s);
    } else {
      String s0 = getStateString(0, s);
      lines.add(s0 + "\n");
    } 
    StringBuilder buf = new StringBuilder(0);
    Collections.sort(lines);
    for (int i = 0; i < lines.size(); i++) {
      String line = lines.get(i);
      buf.append(line);
    } 
    return buf.toString();
  }
  
  protected void walkFANormalizingStateNumbers(State s) {
    if (s == null) {
      ErrorManager.internalError("null state s");
      return;
    } 
    if (this.stateNumberTranslator.get(s) != null)
      return; 
    this.stateNumberTranslator.put(s, Utils.integer(this.stateCounter));
    this.stateCounter++;
    for (int i = 0; i < s.getNumberOfTransitions(); i++) {
      Transition edge = s.transition(i);
      walkFANormalizingStateNumbers(edge.target);
      if (edge instanceof RuleClosureTransition)
        walkFANormalizingStateNumbers(((RuleClosureTransition)edge).followState); 
    } 
  }
  
  protected void walkSerializingFA(List<String> lines, State s) {
    if (this.markedStates.contains(s))
      return; 
    this.markedStates.add(s);
    int normalizedStateNumber = s.stateNumber;
    if (this.stateNumberTranslator != null) {
      Integer normalizedStateNumberI = this.stateNumberTranslator.get(s);
      normalizedStateNumber = normalizedStateNumberI.intValue();
    } 
    String stateStr = getStateString(normalizedStateNumber, s);
    for (int i = 0; i < s.getNumberOfTransitions(); i++) {
      Transition edge = s.transition(i);
      StringBuilder buf = new StringBuilder();
      buf.append(stateStr);
      if (edge.isAction()) {
        buf.append("-{}->");
      } else if (edge.isEpsilon()) {
        buf.append("->");
      } else if (edge.isSemanticPredicate()) {
        buf.append("-{").append(edge.label.getSemanticContext()).append("}?->");
      } else {
        String predsStr = "";
        if (edge.target instanceof DFAState) {
          SemanticContext preds = ((DFAState)edge.target).getGatedPredicatesInNFAConfigurations();
          if (preds != null)
            predsStr = "&&{" + preds.genExpr(this.grammar.generator, this.grammar.generator.getTemplates(), null).render() + "}?"; 
        } 
        buf.append("-").append(edge.label.toString(this.grammar)).append(predsStr).append("->");
      } 
      int normalizedTargetStateNumber = edge.target.stateNumber;
      if (this.stateNumberTranslator != null) {
        Integer normalizedTargetStateNumberI = this.stateNumberTranslator.get(edge.target);
        normalizedTargetStateNumber = normalizedTargetStateNumberI.intValue();
      } 
      buf.append(getStateString(normalizedTargetStateNumber, edge.target));
      buf.append("\n");
      lines.add(buf.toString());
      walkSerializingFA(lines, edge.target);
      if (edge instanceof RuleClosureTransition)
        walkSerializingFA(lines, ((RuleClosureTransition)edge).followState); 
    } 
  }
  
  private String getStateString(int n, State s) {
    String stateStr = ".s" + n;
    if (s.isAcceptState())
      if (s instanceof DFAState) {
        stateStr = ":s" + n + "=>" + ((DFAState)s).getUniquelyPredictedAlt();
      } else {
        stateStr = ":s" + n;
      }  
    return stateStr;
  }
}
